import pathlib
from dataclasses import dataclass
from typing import Any, DefaultDict, Dict, Mapping, Optional, Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    MypyPreset,
    MypySetting,
    MypyTarget,
    PluginBase,
    PluginConfig,
    RunOptions,
    Source,
)
from pysen.component import LintComponentBase
from pysen.ext import mypy_wrapper
from pysen.lint_command import LintCommandBase
from pysen.path import resolve_path
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.setting import SettingFile
from pysen.source import PythonFileFilter
from pysen.types import TargetName

_SettingFileName = "setup.cfg"


def mypy_simple() -> PluginBase:
    return MypySimplePlugin()


def _parse_mypy_preset(s: Any) -> MypyPreset:
    if not isinstance(s, str):
        raise dacite.WrongTypeError(MypyPreset, s)

    try:
        return MypyPreset[s.upper()]
    except KeyError:
        raise dacite.DaciteError(f"invalid mypy_preset value: {s}") from None


@dataclass
class MypySimplePluginConfig:
    preset: Optional[MypyPreset] = None


class MypySimplePlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config_data: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        config = dacite.from_dict(
            MypySimplePluginConfig,
            config_data.config or {},
            dacite.Config(type_hooks={MypyPreset: _parse_mypy_preset}, strict=True),
        )

        if root.lint:
            global_presest = root.lint.mypy_preset or MypyPreset.STRICT
        else:
            global_presest = MypyPreset.STRICT

        preset = config.preset or global_presest

        return [MypySimpleComponent(source, setting=preset.get_setting())]


class MypySimpleComponent(LintComponentBase):
    def __init__(
        self,
        source: Source,
        name: str = "mypy_simple",
        setting: Optional[MypySetting] = None,
        module_settings: Optional[Mapping[str, MypySetting]] = None,
    ) -> None:
        super().__init__(name, source)
        self._setting = setting or MypySetting()
        self._module_settings: Dict[str, MypySetting] = dict(module_settings or {})

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint"]

    @property
    def name(self) -> str:
        return self._name

    @property
    def setting(self) -> MypySetting:
        return self._setting

    @property
    def module_settings(self) -> Dict[str, MypySetting]:
        return self._module_settings

    def export_settings(
        self,
        paths: PathContext,
        files: DefaultDict[str, SettingFile],
    ) -> None:
        setting_file = files[_SettingFileName]
        global_section, global_setting = self._setting.export(paths.base_dir)
        setting_file.set_section(global_section, global_setting)

        for module_name, setting in self._module_settings.items():
            section, module_setting = setting.export(
                paths.base_dir, target_module=module_name
            )
            setting_file.set_section(section, module_setting)

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        if target == "lint":
            return MypySimpleCommand(
                self.name, paths, self.source, options.require_diagnostics
            )

        raise AssertionError(f"unknown {target}")


class MypySimpleCommand(LintCommandBase):
    def __init__(
        self, name: str, paths: PathContext, source: Source, require_diagnostics: bool
    ) -> None:
        super().__init__(paths.base_dir, source)
        self._name = name
        self._setting_path = resolve_path(paths.settings_dir, _SettingFileName)
        self._require_diagnostics = require_diagnostics

    @property
    def name(self) -> str:
        return self._name

    @property
    def has_side_effects(self) -> bool:
        return False

    def __call__(self, reporter: Reporter) -> int:
        sources = self._get_sources(reporter, PythonFileFilter)
        reporter.logger.info(f"Checking {len(sources)} files")
        target = MypyTarget(paths=list(sources))
        return mypy_wrapper.run(
            reporter,
            self.base_dir,
            self._setting_path,
            target,
            self._require_diagnostics,
        )

    def run_files(self, reporter: Reporter, files: Sequence[pathlib.Path]) -> int:
        covered_files = self._get_covered_files(reporter, files, PythonFileFilter)

        if len(covered_files) == 0:
            return 0

        return mypy_wrapper.run(
            reporter,
            self.base_dir,
            self._setting_path,
            MypyTarget(list(covered_files)),
            self._require_diagnostics,
        )



import os
import pathlib
from typing import Iterator, List, Sequence, Tuple

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    Source,
    git_utils,
)
from pysen.mypy import MypyTarget
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName


def mypy_init_check() -> PluginBase:
    return MypyInitCheckPlugin()


class MypyInitCheckPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.mypy_targets:
            mypy_targets = root.lint.mypy_targets
        else:
            mypy_targets = []

        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        return [MypyInitCheckComponent(mypy_targets, source)]


class MypyInitCheckComponent(ComponentBase):
    def __init__(self, mypy_targets: List[MypyTarget], source: Source) -> None:
        self._mypy_targets = mypy_targets
        self._source = source

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return MypyInitCheckCommand(paths.base_dir, self._mypy_targets, self._source)


class MypyInitCheckCommand(CommandBase):
    def __init__(
        self, base_dir: pathlib.Path, mypy_targets: List[MypyTarget], source: Source
    ) -> None:
        self._base_dir = base_dir
        self._mypy_targets = mypy_targets
        self._source = source

    @property
    def name(self) -> str:
        return "mypy_init_check"

    @property
    def has_side_effects(self) -> bool:
        return False

    def __call__(self, reporter: Reporter) -> int:
        targets = set(
            target
            for mypy_target in self._mypy_targets
            for path in mypy_target.paths
            for target in (path.iterdir() if path.is_dir() else (path,))
        )

        sources = self._source.resolve_files(
            self._base_dir,
            lambda file_path: file_path.suffix == ".py",
            git_utils.check_git_available(self._base_dir),
            reporter,
        )

        error = 0
        for source in sources:
            if source in targets:
                continue

            missing_inits = []
            for parent in source.parents:
                init = parent / "__init__.py"
                if not init.is_file():
                    missing_inits.append(init)
                if parent in targets:
                    if len(missing_inits) > 0:
                        error += 1
                        reporter.logger.error(
                            f"{source}: missing __init__.py(s). "
                            f"({', '.join(str(source) for source in reversed(missing_inits))})"
                        )
                    break
            else:
                error += 1
                reporter.logger.error(f"{source}: not a descendant of mypy_targets.")

        return error


def _walk(
    top: pathlib.Path,
) -> Iterator[Tuple[pathlib.Path, List[pathlib.Path], List[pathlib.Path]]]:
    for root, dirs, files in os.walk(top):
        yield (
            pathlib.Path(root),
            [pathlib.Path(dir) for dir in dirs],
            [pathlib.Path(file) for file in files],
        )



import dataclasses
import logging
import pathlib
from typing import Optional, Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    SingleFileFormatCommandBase,
    Source,
    process_utils,
)
from pysen.component import LintComponentBase
from pysen.reporter import Reporter
from pysen.runner_options import PathContext


def cmake_format() -> PluginBase:
    return CMakeFormatPlugin()


class CMakeFormatPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, raw_config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint is not None and root.lint.source is not None:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        config: CMakeFormatConfig
        if raw_config.config is not None:
            config = dacite.from_dict(
                CMakeFormatConfig,
                raw_config.config or {},
                config=dacite.Config(
                    type_hooks={pathlib.Path: lambda p: file_path.parent / p},
                    strict=True,
                ),
            )
            assert isinstance(config, CMakeFormatConfig)
        else:
            config = CMakeFormatConfig()

        return [CMakeFormatComponent(source, config)]


@dataclasses.dataclass
class CMakeFormatConfig:
    config_files: Optional[Sequence[pathlib.Path]] = None


class CMakeFormatComponent(LintComponentBase):
    def __init__(self, source: Source, config: CMakeFormatConfig) -> None:
        super().__init__("cmake_format", source)
        self._config = config

    @property
    def targets(self) -> Sequence[str]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return CMakeFormatCommand(
            paths.base_dir, self.source, self._config, target == "format"
        )


class CMakeFormatCommand(SingleFileFormatCommandBase):
    def __init__(
        self,
        base_dir: pathlib.Path,
        source: Source,
        config: CMakeFormatConfig,
        inplace_edit: bool,
    ) -> None:
        super().__init__(base_dir, source, inplace_edit)
        self._config = config

    @property
    def name(self) -> str:
        return "cmake_format"

    @property
    def has_side_effects(self) -> bool:
        return self.inplace_edit

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.name == "CMakeLists.txt" or file_path.suffix == ".cmake"

    def __call__(self, reporter: Reporter) -> int:
        if self._config.config_files is not None:
            for p in self._config.config_files:
                if not p.exists():
                    raise RuntimeError(f"config_file: {p} does not exist")

        return super().__call__(reporter)

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        cmd = [
            "cmake-format",
        ]
        if self._config.config_files is not None:
            cmd.append("--config-files")
            cmd.extend(map(str, self._config.config_files))

        cmd += [
            "--",
            str(file_path),
        ]
        returncode, stdout, _ = process_utils.run(
            cmd,
            reporter,
            stdout_loglevel=logging.NOTSET,
        )
        if returncode == 0:
            return stdout
        else:
            return None



base_ignore = [
    "E501",  # Line too long (ignored by pysen)
    "D",  # pydocstyle (https://pypi.org/project/pydocstyle/)
    "ICN",  # flake8-import-conventions (https://github.com/joaopalmeiro/flake8-import-conventions) # NOQA
]
strict_ignore = [
    *base_ignore,
    "S101",  # Use of `assert` detected
    "T20",  # flake8-print (https://pypi.org/project/flake8-print/)
    "ANN",  # flkae8-annotations (https://pypi.org/project/flake8-annotations/)
    "INP001",  # File {file} is part of an implicit namespace package. Add an `__init__.py`.
]

basic_select = [
    "I",  # isort (https://pypi.org/project/isort/)
    "E",  # pycodestyle (https://pypi.org/project/pycodestyle/)
    "F",  # pyflakes (https://pypi.org/project/pyflakes/)
    "W",  # pycodestyle (https://pypi.org/project/pycodestyle/)
    "B",  # flake8-bugbear (https://pypi.org/project/flake8-bugbear/)
]

UNFIXABLE = [
    "ERA001",  # do not delete commented code
]

preset_dict = {
    "basic": (basic_select, strict_ignore),
    "strict": (["ALL"], strict_ignore),
    "very_strict": (["ALL"], base_ignore),
    "all": (["ALL"], base_ignore),  # TODO: delete
}



import dataclasses
import logging
import pathlib
from typing import Optional, Sequence

import dacite

from pysen import (
    CommandBase,
    ComponentBase,
    Config,
    PluginBase,
    PluginConfig,
    RunOptions,
    SingleFileFormatCommandBase,
    Source,
    process_utils,
)
from pysen.reporter import Reporter
from pysen.runner_options import PathContext
from pysen.types import TargetName


def tidy() -> PluginBase:
    return TidyPlugin()


class TidyPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint and root.lint.source:
            source = root.lint.source
        else:
            source = Source(includes=["."])

        tidy_config = dacite.from_dict(
            TidyConfig,
            config.config or {},
            config=dacite.Config(
                type_hooks={pathlib.Path: lambda p: file_path.parent / p},
                strict=True,
            ),
        )
        assert isinstance(tidy_config, TidyConfig)

        return [TidyComponent(source, tidy_config)]


@dataclasses.dataclass
class TidyConfig:
    config: pathlib.Path
    extensions: Sequence[str]


class TidyComponent(ComponentBase):
    def __init__(self, source: Source, config: TidyConfig) -> None:
        self._source = source
        self._config = config

    @property
    def targets(self) -> Sequence[TargetName]:
        return ["lint", "format"]

    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        return TidyCommand(
            paths.base_dir, self._source, target == "format", self._config
        )


class TidyCommand(SingleFileFormatCommandBase):
    def __init__(
        self,
        base_dir: pathlib.Path,
        source: Source,
        inplace_edit: bool,
        config: TidyConfig,
    ) -> None:
        super().__init__(base_dir, source, inplace_edit)
        self._config = config

    @property
    def name(self) -> str:
        return "tidy"

    @property
    def has_side_effects(self) -> bool:
        return self.inplace_edit

    def filter(self, file_path: pathlib.Path) -> bool:
        return file_path.suffix in self._config.extensions

    def format(self, file_path: pathlib.Path, reporter: Reporter) -> Optional[str]:
        returncode, stdout, stderr = process_utils.run(
            ("tidy", "-quiet", "-config", str(self._config.config), str(file_path)),
            reporter,
            stdout_loglevel=logging.NOTSET,
            stderr_loglevel=logging.NOTSET,
        )
        if returncode == 0:
            return stdout
        else:
            reporter.process_output.log(logging.WARNING, file_path)
            reporter.process_output.log(logging.WARNING, stderr.rstrip("\n"))
            return None



import pathlib
from typing import Sequence

from pysen import ComponentBase, Config, PluginBase, PluginConfig
from pysen.exceptions import InvalidConfigurationError

from .component import PylintComponent


def pylint() -> PluginBase:
    return PylintPlugin()


class PylintPlugin(PluginBase):
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        if root.lint is not None and root.lint.source is not None:
            source = root.lint.source
        else:
            raise InvalidConfigurationError("Source for pylint must be specified")
        return [PylintComponent(source)]
